defmodule Pleroma.HTTP.Backoff do
  alias Pleroma.HTTP
  require Logger

  @cachex Pleroma.Config.get([:cachex, :provider], Cachex)
  @backoff_cache :http_backoff_cache

  defp maybe_parse_date(value) when is_binary(value) do
    case DateTime.from_iso8601(value) do
      {:ok, stamp, _} -> stamp
      _ -> value
    end
  end

  defp maybe_parse_date(value), do: value

  defp maybe_parse_offset(value) when is_binary(value) do
    case Integer.parse(value) do
      {offset, _} -> DateTime.add(DateTime.utc_now(), offset, :second)
      _ -> value
    end
  end

  defp maybe_parse_offset(value), do: value

  defp date_or_nil(value) do
    case value do
      %DateTime{} -> value
      _ -> nil
    end
  end

  defp parse_reset_header(headers, headername) do
    :proplists.get_value(headername, headers, nil)
    |> maybe_parse_date()
    |> maybe_parse_offset()
    |> date_or_nil()
  end

  defp default_backoff() do
    DateTime.utc_now()
    |> Timex.shift(seconds: 5 * 60)
  end

  defp next_backoff_timestamp(%{headers: headers}) when is_list(headers) do
    # figure out from a 429 or 503 response when we can make the next request
    # mastodon uses the x-ratelimit-reset header, so we will prefer that, others Retry-After
    # or something else entirely. If we can't find a header, default to 5 minutes from now
    header_reset =
      parse_reset_header(headers, "x-ratelimit-reset") ||
        parse_reset_header(headers, "retry-after")

    case header_reset do
      nil ->
        defback = default_backoff()

        Logger.error(
          "Rate limited, but couldn't find timestamp! Using default 5 minute backoff until #{defback}"
        )

        defback

      value ->
        Logger.error("Rate limited until #{value}")
        value
    end
  end

  defp next_backoff_timestamp(_), do: default_backoff()

  def get(url, headers \\ [], options \\ []) do
    # this acts as a single throughput for all GET requests
    # we will check if the host is in the cache, and if it is, we will automatically fail the request
    # this ensures that we don't hammer the server with requests, and instead wait for the backoff to expire
    # this is a very simple implementation, and can be improved upon!
    %{host: host} = URI.parse(url)

    case @cachex.get(@backoff_cache, host) do
      {:ok, nil} ->
        case HTTP.get(url, headers, options) do
          {:ok, env} ->
            case env.status do
              status when status in [429, 503] ->
                Logger.error("Rate limited on #{host}! Backing off...")
                timestamp = next_backoff_timestamp(env)
                ttl = Timex.diff(timestamp, DateTime.utc_now(), :seconds)
                # we will cache the host for 5 minutes
                @cachex.put(@backoff_cache, host, true, ttl: ttl)
                {:error, :ratelimit}

              _ ->
                {:ok, env}
            end

          {:error, env} ->
            {:error, env}
        end

      _ ->
        {:error, :ratelimit}
    end
  end
end
